#include "stdafx.h"
#include "CseXml.h"

//#define USE_FILE_MAPPING

#ifdef USE_FILE_MAPPING
	#include "Utils/FileMapping.h"
#endif

#include <sys/stat.h>
#include <fcntl.h>
#include <io.h>

CXmlTree::CXmlTree()
{
	Open();
}

CXmlTree::CXmlTree(cwstringref wsFilePath, bool bCreateIfNotExist)
{
	Open(wsFilePath, bCreateIfNotExist);
}

bool CXmlTree::IsOpen() const
{
	return m_xml && *m_xml;
}

bool CXmlTree::Open()
{
	m_xml = shared_ptr<XmlNode>(new XmlNode(""));
	return IsOpen();
}

bool CXmlTree::Open(cwstringref wsFilePath, bool bCreateIfNotExist)
{
	if (!wsFilePath.size())
		return Open();

	m_wsPath = wsFilePath;

	#if defined USE_FILE_MAPPING

	Utils::Files::FileMapping file(wsFilePath);

	if (!file.isOpened())
	{
		if (bCreateIfNotExist)
			return Open();
		
		return false;
	}

	Utils::Files::MappedRegion region = file.region();
	if (!region.isValid()) 
		return false;

	lpbyte buf = static_cast<lpbyte>(region.data());
	size_t sz = region.size();

	if (!buf || !sz)
		return false;

	int parsed = (int)sz;
	m_xml = shared_ptr<XmlNode>(new XmlNode((const char*)buf, &parsed, ""));

	file.close();
	region.close();

	#else

	int f = _wopen(wsFilePath.c_str(), O_BINARY|O_RDONLY);

	if (f==-1 || f==0)
	{
		if (bCreateIfNotExist)
			return Open();

		return false;
	}

	size_t sz = _filelength(f);
	shared_ptr<char> buf(new char [sz]);
	_read(f, buf.get(), sz);
	_close(f);

	if (!buf || !sz)
		return false;

	int parsed = (int)sz;
	m_xml = shared_ptr<XmlNode>(new XmlNode((const char*)buf.get(), &parsed, ""));
	
	#endif

	return IsOpen();
}

bool CXmlTree::Close()
{
	m_xml.reset();
	m_wsPath.clear();
	return true;
}

bool CXmlTree::Save(cwstringref wsPath)
{
	if (!GetHandle())
		return false;

	wstring wsFilePath = wsPath.size() ? wsPath : m_wsPath;

	if (!wsFilePath.size())
		return false;

	HANDLE file = ::CreateFileW(wsFilePath.c_str()
		, FILE_WRITE_DATA
		, FILE_SHARE_READ | FILE_SHARE_WRITE
		, 0
		, CREATE_ALWAYS
		, 0
		, 0);

	if (file == INVALID_HANDLE_VALUE)
		return false;

	int len = 0;
	LPSTR szStr = xi.toString(GetHandle(), &len);

	DWORD dw = 0;
	WriteFile(file, szStr, len, &dw, 0);
	CloseHandle(file);
	free(szStr);
	return true;
}

bool CXmlTree::SaveAndClose(cwstringref wsPath)
{
	bool res = Save(wsPath);
	res = res && Close();
	return res;
}

bool CXmlTree::SaveFormattedR(HXML hxml, int level)
{
	if (!hxml)
		return false;

	shared_ptr<char> tab(new char [level]);
	memset(tab.get(), '\x9', level);
	
	int iChilds = xi.getChildCount(hxml);

	string sText;
	LPCSTR pszText = xi.getText(hxml);

	if (pszText)
		sText = pszText;

	boost::algorithm::trim(sText);

	LPCSTR pszName = xi.getName(hxml);

	if (!pszName)
		return false;

	_write(m_file, tab.get(), level);
	_write(m_file, "<", 1);
	_write(m_file, pszName, strlen(pszName));

	int iAttrs = xi.getAttrCount(hxml);

	for (int i=0; i<iAttrs; ++i)
	{
		LPCSTR pszAttrName = xi.getAttrName(hxml, i);
		LPCSTR pszAttrValue = xi.getAttrValue(hxml, pszAttrName);

		_write(m_file, " ", 1);
		_write(m_file, pszAttrName, strlen(pszAttrName));
		_write(m_file, "=\"", 2);
		_write(m_file, pszAttrValue, strlen(pszAttrValue));
		_write(m_file, "\"", 1);
	}

	if (iChilds || sText.find('\n')!=string::npos)
	{
		_write(m_file, ">\r\n", 3);

		if (sText.size())
		{
			_write(m_file, sText.data(), sText.size());
			_write(m_file, "\r\n", 2);
		}

		int iChilds = xi.getChildCount(hxml);
		
		for (int i=0; i<iChilds; ++i)
			SaveFormattedR(xi.getChild(hxml, i), level+1);

		_write(m_file, tab.get(), level);
		_write(m_file, "</", 2);
		_write(m_file, pszName, strlen(pszName));
		_write(m_file, ">\r\n", 3);
	}
	else if (sText.size())
	{
		_write(m_file, ">", 1);
		_write(m_file, sText.data(), sText.size());
		_write(m_file, "</", 2);
		_write(m_file, pszName, strlen(pszName));
		_write(m_file, ">\r\n", 3);
	}
	else
	{
		_write(m_file, "/>\r\n", 4);
	}

	return true;
}

bool CXmlTree::SaveFormatted(cwstringref wsFilePath)
{
	m_file = _wopen(wsFilePath.c_str(), O_CREAT|O_BINARY|O_WRONLY|O_TRUNC, _S_IREAD|_S_IWRITE);
	
	if (m_file==-1 || m_file==0)
		return false;
	
	if (IsOpen())
		SaveFormattedR(xmlGetChild(GetHandle(), 0), 0);

	_close(m_file);

	return true;
}

bool CXmlTree::SaveAndCloseFormatted(cwstringref wsFilePath)
{
	bool res = SaveFormatted(wsFilePath);
	res = res && Close();
	return res;
}
