#ifndef __CseXml_h__
#define __CseXml_h__

#include "xml/jabber_xml.h"
#include "utils/StringUtils.h"
#include <boost/algorithm/string.hpp>

class CXmlNode
{
protected:
	virtual HXML GetHandle() const
	{
		return m_hxml;
	}

protected:
	shared_ptr<CXmlNode> _GetChild(cstringref name)
	{
		if (!GetHandle())
			return shared_ptr<CXmlNode>();

		HXML hxml = ::xmlGetChild(GetHandle(), name.c_str());
		return hxml ? shared_ptr<CXmlNode>(new CXmlNode(hxml)) : shared_ptr<CXmlNode>();
	}

	shared_ptr<CXmlNode> _GetChild(cstringref name, cstringref attr, cstringref value)
	{
		if (!GetHandle())
			return shared_ptr<CXmlNode>();

		HXML hxml = ::xmlGetChildByTag(GetHandle(), name.c_str(), attr.c_str(), value.c_str());
		return hxml ? shared_ptr<CXmlNode>(new CXmlNode(hxml)) : shared_ptr<CXmlNode>();
	}

	shared_ptr<CXmlNode> _GetaChild(cstringref name)
	{
		if (!GetHandle())
			return shared_ptr<CXmlNode>();

		shared_ptr<CXmlNode> node = _GetChild(name);
		return node ? node : AddChild(name);
	}

	shared_ptr<CXmlNode> _GetaChild(cstringref name, cstringref attr, cstringref value)
	{
		if (!GetHandle())
			return shared_ptr<CXmlNode>();

		shared_ptr<CXmlNode> node = _GetChild(name, attr, value);
		return node ? node : AddChild(name, attr, value);
	}

public:
	CXmlNode(HXML hxml = 0):m_hxml(hxml)
	{

	}

	bool IsOk() const
	{
		return GetHandle() != 0;
	}

	string GetName() const
	{
		if (!GetHandle())
			return string();

		LPCSTR pszName = ::xmlGetName(GetHandle());
		return pszName ? pszName : string();
	}

	string GetText() const
	{
		if (!GetHandle())
			return string();

		LPCSTR pszText = ::xmlGetText(GetHandle());
		return pszText ? pszText : string();
	}

	int GetChildCount() const
	{
		return GetHandle() ? ::xmlGetChildCount(GetHandle()) : 0;
	}

	shared_ptr<CXmlNode> GetChild(int index)
	{
		if (!GetHandle())
			return shared_ptr<CXmlNode>();

		HXML hxml = ::xmlGetChild(GetHandle(), index);
		return hxml ? shared_ptr<CXmlNode>(new CXmlNode(hxml)) : shared_ptr<CXmlNode>();
	}

	shared_ptr<CXmlNode> AddChild(cstringref name)
	{
		if (!GetHandle())
			return shared_ptr<CXmlNode>();

		HXML hxml = ::xmlAddChild(GetHandle(), name.c_str());
		return hxml ? shared_ptr<CXmlNode>(new CXmlNode(hxml)) : shared_ptr<CXmlNode>();
	}

	shared_ptr<CXmlNode> AddChild(cstringref name, cstringref attr, cstringref value)
	{
		shared_ptr<CXmlNode> child = AddChild(name);

		if (child)
			child->SetaAttr(attr, value);

		return child;
	}

	shared_ptr<CXmlNode> GetNode(cstringref path)
	{
		//<domains><domain name="dd.net"><routes><route host="192.168.0.1"></routes></domain></domains>
		//"domains/domain@name='dd.net'/routes/route@host='192.168.0.1'"

		if (!GetHandle())
			return shared_ptr<CXmlNode>();

		Utils::Splitter split(path, '/');

		if (!split.size())
			return shared_ptr<CXmlNode>();

		shared_ptr<CXmlNode> node(new CXmlNode(GetHandle()));

		for (size_t i=0; i<split.size(); ++i)
		{
			Utils::Splitter sub(split[i], '@');

			if (sub.size() == 1)
			{
				node = node->_GetChild(sub[0]);
			}
			else
			{
				if (sub.size() > 2)
					return shared_ptr<CXmlNode>();

				Utils::Splitter attr(sub[1], '=');

				if (attr.size() != 2)
					return shared_ptr<CXmlNode>();

				if (attr[1].size()<2 || attr[1][0]!='\'' || attr[1][attr[1].size()-1]!='\'')
					return shared_ptr<CXmlNode>();

				node = node->_GetChild(sub[0], attr[0], attr[1].substr(1, attr[1].size()-2));
			}

			if (!node)
				return node;
		}

		return node;
	}

	shared_ptr<CXmlNode> GetaNode(cstringref path)
	{
		//<domains><domain name="dd.net"><routes><route host="192.168.0.1"></routes></domain></domains>
		//"domains/domain@host='dd.net'/routes/route@host='192.168.0.1'"

		if (!GetHandle())
			return shared_ptr<CXmlNode>();

		Utils::Splitter split(path, '/');

		if (!split.size())
			return shared_ptr<CXmlNode>();

		shared_ptr<CXmlNode> node(new CXmlNode(GetHandle()));

		for (size_t i=0; i<split.size(); ++i)
		{
			Utils::Splitter sub(split[i], '@');

			if (sub.size() == 1)
			{
				node = node->_GetaChild(sub[0]);
			}
			else
			{
				if (sub.size() > 2)
					return shared_ptr<CXmlNode>();

				Utils::Splitter attr(sub[1], '=');

				if (attr.size() != 2)
					return shared_ptr<CXmlNode>();

				if (attr[1].size()<2 || attr[1][0]!='\'' || attr[1][attr[1].size()-1]!='\'')
					return shared_ptr<CXmlNode>();

				node = node->_GetaChild(sub[0], attr[0], attr[1].substr(1, attr[1].size()-2));
			}

			if (!node)
				return node;
		}

		return node;
	}

	bool DeleteChild(int index)
	{
		if (index < GetChildCount())
		{
			HXML hxml = ::xmlGetChild(GetHandle(), index);

			if (hxml)
			{
				xi.destroyNode(hxml);
				return true;
			}
		}

		return false;
	}

	bool DeleteChild(shared_ptr<CXmlNode> node)
	{
		if (!node)
			return false;

		for (int index=0; index<GetChildCount(); ++index)
		{
			if (node->GetHandle() == ::xmlGetChild(GetHandle(), index))
			{
				xi.destroyNode(node->GetHandle());
				return true;
			}
		}

		return false;
	}

	bool DeleteChild(cstringref name)
	{
		shared_ptr<CXmlNode> child = _GetChild(name);
		return child ? DeleteChild(child) : false;
	}

	bool DeleteChild(cstringref name, cstringref attr, cstringref value)
	{
		shared_ptr<CXmlNode> child = _GetChild(name, attr, value);
		return child ? DeleteChild(child) : false;
	}

	bool IsAttrExist(cstringref name)
	{
		return ::xi.getAttrValue(GetHandle(), name.c_str()) != 0;
	}

	bool AddAttr(cstringref name, cstringref value)
	{
		::xi.addAttr(GetHandle(), name.c_str(), value.c_str());
		return true;
	}

	bool SetAttr(cstringref name, cstringref value)
	{
		return ::xi.setAttrValue(GetHandle(), name.c_str(), value.c_str()) != 0;
	}

	bool SetaAttr(cstringref name, cstringref value)
	{
		return SetAttr(name, value) ? true : AddAttr(name, value);
	}

	int GetAttrCount() const
	{
		return ::xi.getAttrCount(GetHandle());
	}

	string GetAttrName(int index) const
	{
		LPCSTR psz = ::xi.getAttrName(GetHandle(), index);
		return psz ? string(psz) : string();
	}

	string GetAttrValue(cstringref name) const
	{
		LPCSTR psz = ::xi.getAttrValue(GetHandle(), name.c_str());
		return psz ? string(psz) : string();
	}

	//////////////////////////////////////////////////////////////////////////

	string GetAttrString(cstringref name) const
	{
		return GetAttrValue(name);
	}

	bool SetAttrString(cstringref name, cstringref str)
	{
		return SetAttr(name, str);
	}

	bool SetaAttrString(cstringref name, cstringref str)
	{
		return SetaAttr(name, str);
	}

	bool SetAttrStringList(cstringref name, const std::list<string>& strs)
	{
		string str;
		
		for(list<string>::const_iterator it=strs.begin(); it!=strs.end(); ++it)
		{
			if (it != strs.begin())
				str+=";";

			str += *it;
		}

		return SetAttr(name, str);
	}

	bool SetaAttrStringList(cstringref name, const std::list<string>& strs)
	{
		string str;

		for(list<string>::const_iterator it=strs.begin(); it!=strs.end(); ++it)
		{
			if (it != strs.begin())
				str+=";";

			str += *it;
		}

		return SetaAttr(name, str);
	}

	int32 GetAttrInt32(cstringref name) const
	{
		return Utils::lex_cast<int32>(GetAttrString(name));
	}

	void SetAttrInt32(cstringref name, int32 v)
	{
		SetAttrString(name, Utils::lex_cast<string>(v));
	}

	void SetaAttrInt32(cstringref name, int32 v)
	{
		SetaAttrString(name, Utils::lex_cast<string>(v));
	}

	uint32 GetAttrUInt32(cstringref name) const
	{
		return Utils::lex_cast<uint32>(GetAttrString(name));
	}

	void SetAttrUInt32(cstringref name, uint32 v)
	{
		SetAttrString(name, Utils::lex_cast<string>(v));
	}

	void SetaAttrUInt32(cstringref name, uint32 v)
	{
		SetaAttrString(name, Utils::lex_cast<string>(v));
	}

	int64 GetAttrInt64(cstringref name) const
	{
		return Utils::lex_cast<int64>(GetAttrString(name));
	}

	void SetAttrInt64(cstringref name, int64 v)
	{
		SetAttrString(name, Utils::lex_cast<string>(v));
	}

	void SetaAttrInt64(cstringref name, int64 v)
	{
		SetaAttrString(name, Utils::lex_cast<string>(v));
	}

	uint64 GetAttrUInt64(cstringref name) const
	{
		return Utils::lex_cast<uint64>(GetAttrString(name));
	}

	void SetAttrUInt64(cstringref name, uint64 v)
	{
		SetAttrString(name, Utils::lex_cast<string>(v));
	}

	void SetaAttrUInt64(cstringref name, uint64 v)
	{
		SetaAttrString(name, Utils::lex_cast<string>(v));
	}

	bool GetAttrBool(cstringref name) const
	{
		string s = boost::to_lower_copy(GetAttrString(name));

		if (s=="true" || s=="on" || s=="enabled" || s=="active")
			return true;

		try
		{
			return boost::lexical_cast<int>(s)!=0;
		}
		catch (...)
		{
		}

		return false;
	}

	void SetAttrBool(cstringref name, bool v)
	{
		SetAttrString(name, v ? "true" : "false");
	}

	void SetaAttrBool(cstringref name, bool v)
	{
		SetaAttrString(name, v ? "true" : "false");
	}

private:
	HXML m_hxml;
};

typedef shared_ptr<CXmlNode> SPXmlNode;

class CXmlTree : public CXmlNode
{
protected:
	virtual HXML GetHandle() const
	{
		return m_xml ? *m_xml : 0;
	}

	bool SaveFormattedR(HXML node, int level);

public:
	CXmlTree();
	CXmlTree(cwstringref wsFilePath, bool bCreateIfNotExist = false);

	bool IsOpen() const;
	bool Open();
	bool Open(cwstringref wsFilePath, bool bCreateIfNotExist = false);
	bool Save(cwstringref wsFilePath = wstring());
	bool SaveAndClose(cwstringref wsFilePath = wstring());
	bool SaveFormatted(cwstringref wsFilePath = wstring());
	bool SaveAndCloseFormatted(cwstringref wsFilePath = wstring());
	bool Close();

private:
	shared_ptr<XmlNode>		m_xml;
	wstring					m_wsPath;
	int						m_file;
};

typedef shared_ptr<CXmlTree> SPXmlTree;

#endif // __CseXml_h__
