#ifndef __StringUtils_h__
#define __StringUtils_h__

#include <boost\shared_ptr.hpp>
#include <string>
#include <vector>
#include <stdarg.h>
#include <set>
#include <boost/lexical_cast.hpp>

namespace Utils
{
	template <class T> class strt : public std::basic_string<T>
	{
	public:
		typedef typename std::basic_string<T> xstr;
		typedef typename xstr::const_pointer const_pointer;
		typedef typename xstr::const_iterator const_iterator;
		typedef typename xstr::value_type value_type;
		typedef typename xstr::size_type size_type;
	
	public:
		//operator stdstring() { return ; }

		strt():xstr() {}
		strt(const value_type* _Ptr, size_type _Count):xstr(_Ptr, _Count) {}
		strt(const value_type* _Ptr):xstr(_Ptr) {}
		strt(const xstr& _Right):xstr(_Right) {}
		strt(const xstr& _Right, size_type _Roff, size_type _Count = npos):xstr(_Right, _Roff, _Count) {}
		strt(size_type _Count, value_type _Ch):xstr(_Count, _Ch) {}
		template<class _It> strt(_It _First, _It _Last):xstr(_First, _Last) {}
		strt(const_pointer _First, const_pointer _Last):xstr(_First, _Last) {}
		strt(const_iterator _First, const_iterator _Last):xstr(_First, _Last) {}

		bool ciequal(const value_type* s)
		{
			return boost::to_upper_copy(*this) == boost::to_upper_copy(xstr(s));
		}

		bool ciequal(const strt& s)
		{
			return boost::to_upper_copy(*this) == boost::to_upper_copy(s);
		}

		bool ciequal(const xstr& s)
		{
			return boost::to_upper_copy(*this) == boost::to_upper_copy(s);
		}

		int replace(const xstr& f, const xstr& t)
		{
			if (!f.size())
				return 0;

			if (t.size() > f.size())
			{
				size_t newsize = size(), dif = t.size() - f.size();
				for (size_t i=find(f); i!=npos; i=find(f, i+f.size()), newsize+=dif);
				reserve(newsize);
			}

			int count = 0;
			for (size_t i=find(f); i!=npos; xstr::replace(i, f.size(), t), i=find(f, i+t.size()), ++count);
			return count;
		}
	};

	class bstr : public strt<char>
	{
	public:
		typedef strt<char> xstrt;
		typedef xstrt::xstr xstr;
		typedef xstrt::const_pointer const_pointer;
		typedef xstrt::const_iterator const_iterator;
		typedef xstrt::value_type value_type;
		typedef xstrt::size_type size_type;

	public:
		bstr():xstrt() {}
		bstr(const value_type* _Ptr, size_type _Count):xstrt(_Ptr, _Count) {}
		bstr(const value_type* _Ptr):xstrt(_Ptr) {}
		bstr(const xstr& _Right):xstrt(_Right) {}
		bstr(const xstr& _Right, size_type _Roff, size_type _Count = npos):xstrt(_Right, _Roff, _Count) {}
		bstr(size_type _Count, value_type _Ch):xstrt(_Count, _Ch) {}
		template<class _It>bstr(_It _First, _It _Last):xstrt(_First, _Last) {}
		bstr(const_pointer _First, const_pointer _Last):xstrt(_First, _Last) {}
		bstr(const_iterator _First, const_iterator _Last):xstrt(_First, _Last) {}
		bstr(const bstr& bs):xstrt((xstrt)bs) {}
		bstr(const bstr& _Right, size_type _Roff, size_type _Count = npos):xstrt((xstrt)_Right, _Roff, _Count) {}

		strt& format(const value_type* _Ptr, ...)
		{
			va_list lst;
			va_start(lst, _Ptr);
			int sz = _vscprintf(_Ptr, lst);
			reserve(sz);
			resize(sz);
			sz = _vsnprintf((value_type*)data(), size(), _Ptr, lst);
			va_end(lst);

			if (sz <= 0)
				resize(0);
			else if (sz != size())
				resize(sz);

			return *this;
		}
	};

	class wstr : public strt<wchar_t>
	{
	public:
		typedef strt<wchar_t> xstrt;
		typedef xstrt::xstr xstr;
		typedef xstrt::const_pointer const_pointer;
		typedef xstrt::const_iterator const_iterator;
		typedef xstrt::value_type value_type;
		typedef xstrt::size_type size_type;

	public:
		wstr():xstrt() {}
		wstr(const value_type* _Ptr, size_type _Count):xstrt(_Ptr, _Count) {}
		wstr(const value_type* _Ptr):xstrt(_Ptr) {}
		wstr(const xstr& _Right):xstrt(_Right) {}
		wstr(const xstr& _Right, size_type _Roff, size_type _Count = npos):xstrt(_Right, _Roff, _Count) {}
		wstr(size_type _Count, value_type _Ch):xstrt(_Count, _Ch) {}
		template<class _It>wstr(_It _First, _It _Last):xstrt(_First, _Last) {}
		wstr(const_pointer _First, const_pointer _Last):xstrt(_First, _Last) {}
		wstr(const_iterator _First, const_iterator _Last):xstrt(_First, _Last) {}
		wstr(const wstr& bs):xstrt((xstrt)bs) {}
		wstr(const wstr& _Right, size_type _Roff, size_type _Count = npos):xstrt((xstrt)_Right, _Roff, _Count) {}

		strt& format(const value_type* _Ptr, ...)
		{
			va_list lst;
			va_start(lst, _Ptr);
			int sz = _vscwprintf(_Ptr, lst);
			reserve(sz);
			resize(sz);
			_vsnwprintf((value_type*)data(), size(), _Ptr, lst);
			va_end(lst);
			
			if (sz <= 0)
				resize(0);
			else if (sz != size())
				resize(sz);
			
			return *this;
		}
	};

	#ifdef _UNICODE
	typedef wstr tstr;
	#else
	typedef bstr tstr;
	#endif

	//////////////////////////////////////////////////////////////////////////

	template <class T> int _replace(std::basic_string<T>& s, const std::basic_string<T>& f, const std::basic_string<T>& t);

	int Replace(std::string& s, const std::string& f, const std::string& t);
	int Replace(std::wstring& s, const std::wstring& f, const std::wstring& t);

	template <class T> class SplitterTempl
	{
	public:
		typedef typename std::basic_string<T> str_type;
		typedef str_type value_type;
		typedef typename std::vector<str_type> Result;
		typedef typename boost::shared_ptr<Result> SpResult;
		typedef typename Result::iterator iterator;
		typedef typename Result::const_iterator const_iterator;
		typedef typename Result::reverse_iterator reverse_iterator;
		typedef typename Result::const_reverse_iterator const_reverse_iterator;
		typedef typename str_type::value_type char_type;
		typedef typename str_type::size_type size_type;

	private:
		void _split(std::vector<std::basic_string<T> >& strings, const std::basic_string<T>& str, T sep, bool bMultipleSep, bool bAdd, bool bNoEmpty, bool bTrimEnd)
		{
			if (!bAdd)
				strings.clear();

			if (!str.size())
				return;

			size_t beg = 0, pos = 0;

			while (pos < str.size())
			{
				if (str[pos]!=sep)
					++pos;
				else
				{
					if (!bNoEmpty || beg<pos)
						strings.push_back(std::basic_string<T>(str.data()+beg, pos-beg));

					beg = ++pos;

					if (bMultipleSep)
						for (; str[pos]==sep; beg=++pos);
				}
			}

			if ((!bTrimEnd || pos!=beg) && (!bNoEmpty || beg<pos))
				strings.push_back(std::basic_string<T>(str.data()+beg, pos-beg));
		}

		void _split(std::vector<std::basic_string<T> >& strings, const std::basic_string<T>& str, const std::basic_string<T>& sep, bool bMultipleSep, bool bAdd, bool bNoEmpty, bool bTrimEnd)
		{
			if (!bAdd)
				strings.clear();

			if (!str.size())
				return;

			size_t beg = 0, pos = 0;

			while (pos < str.size())
			{
				if (sep.find(str[pos]) == std::basic_string<T>::npos)
					++pos;
				else
				{
					if (!bNoEmpty || beg<pos)
						strings.push_back(std::basic_string<T>(str.data()+beg, pos-beg));

					beg = ++pos;

					if (bMultipleSep)
						for (; sep.find(str[pos])!=std::basic_string<T>::npos; beg=++pos);
				}
			}

			if ((!bTrimEnd || pos!=beg) && (!bNoEmpty || beg<pos))
				strings.push_back(std::basic_string<T>(str.data()+beg, pos-beg));
		}

		void _split_str_sep(std::vector<std::basic_string<T> >& strings, const std::basic_string<T>& str, const std::basic_string<T>& sep, bool bMultipleSep, bool bAdd, bool bNoEmpty, bool bTrimEnd)
		{
			if (!bAdd)
				strings.clear();

			if (!str.size())
				return;

			size_t beg = 0, pos = str.find(sep);

			for (; pos!=std::basic_string<T>::npos; pos=str.find(sep, beg))
			{
				if (!bNoEmpty || beg<pos)
					strings.push_back(std::basic_string<T>(str.data()+beg, pos-beg));

				beg = pos + sep.size();

				if (bMultipleSep)
					for (; str.substr(beg, sep.size())==sep; beg+=sep.size());
			}

			if ((!bTrimEnd || pos!=str.size()) && (!bNoEmpty || beg<str.size()))
				strings.push_back(str.substr(beg));
		}

	public:
		SplitterTempl():m_spResult(new Result)
		{

		}

		SplitterTempl(const str_type& str, char_type sep, bool bMultiple = false, bool bNoEmpty = false, bool bTrimEnd = true):m_spResult(new Result)
		{
			Split(str, sep, bMultiple, true, bNoEmpty, bTrimEnd);
		}

		SplitterTempl(const str_type& str, const str_type& sep, bool bMultiple = false, bool bNoEmpty = false, bool bTrimEnd = true, bool bWholeSep = false):m_spResult(new Result)
		{
			Split(str, sep, bMultiple, true, bNoEmpty, bTrimEnd, bWholeSep);
		}

		SplitterTempl& Split(const str_type& str, char_type sep, bool bMultiple = false, bool bAdd = false, bool bNoEmpty = false, bool bTrimEnd = true)
		{
			_split(*m_spResult, str, sep, bMultiple, bAdd, bNoEmpty, bTrimEnd);
			return *this;
		}

		SplitterTempl& Split(const str_type& str, const str_type& sep, bool bMultiple = false, bool bAdd = false, bool bNoEmpty = false, bool bTrimEnd = true, bool bWholeSep = false)
		{
			if (bWholeSep)
				_split_str_sep(*m_spResult, str, sep, bMultiple, bAdd, bNoEmpty, bTrimEnd);
			else
				_split(*m_spResult, str, sep, bMultiple, bAdd, bNoEmpty, bTrimEnd);

			return *this;
		}

		const str_type& operator [] (unsigned int i) const
		{
			return (*m_spResult)[i];
		}

		str_type& operator [] (unsigned int i)
		{
			return (*m_spResult)[i];
		}

		size_type size() const
		{
			return m_spResult->size();
		}

		iterator begin()
		{
			return m_spResult->begin();
		}

		const_iterator begin() const
		{
			return m_spResult->begin();
		}

		iterator end()
		{
			return m_spResult->end();
		}

		const_iterator end() const
		{
			return m_spResult->end();
		}

		iterator erase(iterator f, iterator e)
		{
			return m_spResult->erase(f, e);
		}

		reverse_iterator rbegin()
		{
			return m_spResult->rbegin();
		}

		const_reverse_iterator rbegin() const
		{
			return m_spResult->rbegin();
		}

		reverse_iterator rend()
		{
			return m_spResult->rend();
		}

		const_reverse_iterator rend() const
		{
			return m_spResult->rend();
		}

		operator SpResult()
		{
			return m_spResult;
		}

		operator std::vector<str_type> () const
		{
			return m_spResult ? *m_spResult : std::vector<str_type>();
		}

		operator std::list<str_type> () const
		{
			std::list<str_type> res;

			if (!m_spResult)
				return res;

			for (Result::const_iterator it=m_spResult->begin(); it!=m_spResult->end(); ++it)
				res.push_back(*it);

			return res;
		}

		operator std::set<str_type> () const
		{
			std::set<str_type> res;

			if (!m_spResult)
				return res;

			for (Result::const_iterator it=m_spResult->begin(); it!=m_spResult->end(); ++it)
				res.insert(*it);

			return res;
		}

		operator bool () const
		{
			return size()!=0;
		}

	private:
		SpResult	m_spResult;
	};

	typedef SplitterTempl<char> Splitter;
	typedef SplitterTempl<wchar_t> WSplitter;
	typedef SplitterTempl<TCHAR> TSplitter;

	//////////////////////////////////////////////////////////////////////////

	std::string formatstr(const char* str, ...);
	std::wstring wformatstr(const wchar_t* str, ...);

	#ifdef UNICODE
	#define tformatstr wformatstr
	#else
	#define tformatstr formatstr
	#endif

	std::string wcs2mbs(const std::wstring& ws);
	std::wstring mbs2wcs(const std::string& s);

	//////////////////////////////////////////////////////////////////////////

	bool WildcardMatchNoCase(const std::string &sWildcard, const std::string &sString);
	bool WildcardMatch(const std::string &sWildcard, const std::string &sString);
	std::string encodexml(const std::string &str);
	std::string decodexml(const std::string &str);

	cstringref getValue(mapstring map, cstringref key);

	std::string FromADGuid(cstringref guid);

	template<typename Target, typename Source> inline Target lex_cast(const Source &arg)
	{
		try
		{
			return boost::lexical_cast<Target>(arg);
		}
		catch (...)
		{

		}

		return Target();
	}
}

bool xisspace(char ch);

#endif // __StringUtils_h__
