#include <string.h>
#include <string>
#include <vector>
#include <time.h>
 
 
#define HASH_SIZE  65536
#define CACHE_SIZE 32
#define ALLOC_GRAN 8
#define ALLOC_SIZE 1024
 
typedef unsigned int uint32;
 
 
class static_string
{
public:
 
    static size_t get_str_num() throw()
    {
        return strnum;
    }
 
    static size_t get_str_memory() throw()
    {
        return strmemory;
    }
 
    uint32 size() const throw()
    {
        return pointer->length;
    }
 
    uint32 hash() const throw()
    {
        return pointer->hash;
    }
 
    struct strptr
    {
        strptr *next;
        uint32  refcounter;
        uint32  length;
        uint32  hash;
        char    body[4];
    };
   
    struct list_node
    {
        list_node *next;
    };
 
protected:
    strptr *pointer;
 
    static strptr     empty;
    static size_t     strnum;
    static size_t     strmemory;
    static strptr    *table[HASH_SIZE];
    static list_node *pointers[CACHE_SIZE];
    static list_node *allocs;
   
    strptr *alloc_string( uint32 length ) throw()
    {
 
        uint32 alloc = sizeof( strptr ) + length - 3;
        uint32 p = ( alloc - 1 ) / ALLOC_GRAN + 1;
 
        if( p >= CACHE_SIZE )
        {
            return (strptr *)new char[alloc];
        }
 
        alloc = p * ALLOC_GRAN;
 
        if( pointers[p] == 0 )
        {
            char *data = new char[ALLOC_SIZE];
 
            strmemory += ALLOC_SIZE;
 
            list_node *chunk = (list_node *)( data );
            chunk->next = allocs;
            allocs = chunk;
 
            for( size_t i = sizeof( list_node ); i + alloc <= ALLOC_SIZE; i += alloc )
            {
                list_node *result = (list_node *)( data + i );
                result->next = pointers[p];
                pointers[p] = result;
            }
        }
 
        list_node *result = pointers[p];
        pointers[p] = result->next;
        return (strptr *)result;
    }
 
    void release_string( strptr *ptr ) throw()
    {
        uint32 alloc = sizeof( strptr ) + ptr->length - 3;
        uint32 p = ( alloc - 1 ) / ALLOC_GRAN + 1;
 
        if( p >= CACHE_SIZE )
        {   
            delete[] (char *)ptr;
            return;
        }
 
        list_node *result = (list_node *)ptr;
        result->next = pointers[p];
        pointers[p] = result;
    }
 
    void __forceinline str_hash( const char *str, uint32 &hash, uint32 &length ) throw()
    {       
        uint32 res = 0;
        uint32 i = 0;
 
        while( 1 )
        {
            uint32 v = str[i];
            res = res * 5 + v;
            if( v == 0 )
            {
                hash = res;
                length = i;
                return;
            }
            ++i;
        };
    }
 
    void clear() throw()
    {
        uint32 position = pointer->hash % HASH_SIZE;
        strptr *entry = table[position];
        strptr **prev = &table[position];
        while( entry )
        {
            if( entry == pointer )
            {
                *prev = pointer->next;
                release_string( pointer );
 
                if( --strnum == 0 )
                {
                    while( allocs )
                    {
                        list_node *next = allocs->next;
                        delete[] (char *)allocs;
                        allocs = next;
                        strmemory -= ALLOC_SIZE;
                    }
                   
                    for( size_t i = 0; i < CACHE_SIZE; ++i )
                    {
                        pointers[i] = 0;
                    }
                }
 
                return;
            }
            prev = &entry->next;
            entry = entry->next;
        }
    }
 
public:
 
    const char *c_str() throw()
    {
        return &pointer->body[0];
    }
 
    bool operator == ( const static_string &other ) const throw()
    {
        return pointer == other.pointer;
    }
 
    bool operator != ( const static_string &other ) const throw()
    {
        return pointer != other.pointer;
    }
 
    static_string &operator = ( const static_string &other ) throw()
    {
        static_string swp( other );
        strptr *temp = swp.pointer;
        swp.pointer = pointer;
        pointer = temp;
        return *this;
    }
 
    static_string()  throw()
    {
        pointer = &empty;
        pointer->refcounter++;
    }
 
    static_string( const static_string &other  ) throw()
    {
        pointer = other.pointer;
        pointer->refcounter++;
    }
 
    ~static_string( )  throw()
    {
        if( --pointer->refcounter == 0 )
        {
            clear();
        }
    }
 
    static_string( const char *buff ) throw()
    {
        uint32 hash, length;
 
        if( buff == 0 || buff[0] == 0 )
        {
            pointer = &empty;
            pointer->refcounter++;
            return;
        }
 
        str_hash( buff, hash, length );
 
        uint32 position = hash % HASH_SIZE;
        strptr *entry = table[position];
        while( entry )
        {
            if( entry->length == length && entry->hash == hash && !memcmp( entry->body, buff, length ) )
            {
                pointer = entry;
                ++entry->refcounter;
                return;
            }
            entry = entry->next;
        }
 
        ++strnum;
       
        pointer = alloc_string( length );
        char *buffer = &pointer->body[0];
        memcpy( buffer, buff, length + 1 );
        pointer->hash = hash;
        pointer->next = table[position];
        pointer->refcounter = 1;
        pointer->length = length;
        table[position] = pointer;
    }
};
 
size_t  static_string::strnum = 0;
size_t  static_string::strmemory = 0;
static_string::strptr  static_string::empty = { 0, 1, 0, 0, 0, 0, 0, 0 };
static_string::strptr *static_string::table[HASH_SIZE] = {0};
static_string::list_node *static_string::pointers[CACHE_SIZE] = {0};
static_string::list_node *static_string::allocs = 0;
 
const char  *names[] =
{
            "Hip",
            "Thigh_R",
            "ClothRight_1",
            "ClothRight_2",
            "ClothFront_2",
            "ClothFront_3",
            "ClothFront_1",
            "Shoulder_R",
            "Forearm_R",
            "Hand_R",
            "Chest",
            "Cavicle_R",
            "Spine",
            "Foot_R",
            "Head",
            "Tail_1",
            "Shank_R",
            "Stomach",
            "Tail_2",
            "Thigh_L",
            "Jaw",
            "Finger1_1R",
            "Finger1_2R",
            "Finger2_1R",
            "Finger3_1R",
            "Finger4_1R",
            "Finger2_2R",
            "Finger4_1R",
            "Finger4_2L",
            "Finger3_2R",
            "Eyebrow_R",
            "Lip3_R",
            "Eyelid_R",
            "Ear_R",
            "Lip2_R",
            "Lip1_R",
            "Lip1_L",
            "Slot_Mouth_Tongue_2",
            "Slot_Mouth",
            "ClothLeft_1",
            "ClothLeft_2",
            "Shoulder_L",
            "Forearm_L",
            "Hand_L",
            "Cavicle_L",
            "Foot_L",
            "Shank_L",
            "Finger1_1L",
            "Finger1_2L",
            "Finger3_1L",
            "Finger4_1L",
            "Finger2_1L",
            "Finger2_2L",
            "Finger4_1L",
            "Finger4_2L",
            "Finger3_2L",
            "Eyebrow_L",
            "Lip3_L",
            "Eyelid_L",
            "Ear_L",
            "Lip2_L",
            "Slot_Eye_L",
            "Slot_Eye_R",
            "Slot_Mouth",
            "Slot_Head",
            "Slot_Hand_L",
            "Slot_Hand_R",
            "Slot_FX",
            "Slot_BodyFX",
            "Slot_TopFX",
            "Slot_Global",
            "Global",
            "VisualSceneNode"
};
 
template<class X> class Skeleton
{
    enum Sizes
    {
        Size = sizeof( names ) / sizeof ( names[0] )
    };
    X _names[Size];
public:
    Skeleton()
    {
        for( size_t i = 0; i < Size; ++i )
        {
            _names[i] = X( names[i] );
        }
    };
 
    static void Test()
    {
        float cl1 = (float)clock();
        for( size_t i = 0; i < 20; ++i )
        {
            std::vector< Skeleton<X> > s;
            s.resize( 10000 );
        }
        float cl2 = (float)clock();
        float time = ( cl2 - cl1 ) / float( CLOCKS_PER_SEC );
        printf( "time %f for class %s \n",  time, typeid( X ).name() );
    }
};
/* 
int main()
{
    Skeleton<std::string>::Test();
    Skeleton<static_string>::Test();
    return 0;
};
*/