/*

Jabber Protocol Plugin for Miranda IM
Copyright ( C ) 2002-04  Santithorn Bunchua
Copyright ( C ) 2005-09  George Hazan
Copyright ( C ) 2007     Maxim Mluhov

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or ( at your option ) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

File name      : $URL: https://miranda.svn.sourceforge.net/svnroot/miranda/trunk/miranda/protocols/JabberG/jabber_xml.cpp $
Revision       : $Revision: 8779 $
Last change on : $Date: 2009-01-07 04:16:38 +0600 (Ср, 07 янв 2009) $
Last change by : $Author: ghazan $

*/
#include "stdafx.h"
#include "jabber_xml.h"
#include <atlconv.h>

#define TAG_MAX_LEN 128
#define ATTR_MAX_LEN 8192

XML_API xi;

/////////////////////////////////////////////////////////////////////////////////////////
// XmlNode class members

XmlNode::XmlNode( LPCSTR pszName )
{
	m_hXml = xi.createNode( pszName, NULL, 0 );
}

XmlNode::XmlNode( LPCSTR pszName, LPCSTR ptszText )
{
	m_hXml = xi.createNode( pszName, ptszText, 0 );
}

XmlNode::XmlNode( const XmlNode& n )
{
	m_hXml = xi.copyNode( n );
}

XmlNode& XmlNode::operator =( const XmlNode& n )
{	
	if ( m_hXml )
		xi.destroyNode( m_hXml );
	m_hXml = xi.copyNode( n );
	return *this;
}

XmlNode::~XmlNode()
{
	if ( m_hXml ) {
		xi.destroyNode( m_hXml );
		m_hXml = NULL;
}	}

/////////////////////////////////////////////////////////////////////////////////////////

HXML __fastcall operator<<( HXML node, const XCHILDNS& child )
{
	HXML res = xmlAddChild( node, child.name );
	xmlAddAttr( res, "xmlns", child.ns );
	return res;
}

HXML __fastcall operator<<( HXML node, const XQUERY& child )
{
	HXML n = xmlAddChild( node, "query");
	if ( n )
		xmlAddAttr( n, "xmlns", child.ns );
	return n;
}

/////////////////////////////////////////////////////////////////////////////////////////

void __fastcall xmlAddAttr( HXML hXml, LPCSTR name, LPCSTR value )
{
	if ( value )
		xi.addAttr( hXml, name, value );
}

void __fastcall xmlAddAttr( HXML hXml, LPCSTR pszName, int value )
{
	xi.addAttrInt( hXml, pszName, value );
}

void __fastcall xmlAddAttrID( HXML hXml, int id )
{
	char text[ 100 ];
	xmlAddAttr( hXml, "id", text );
}

/////////////////////////////////////////////////////////////////////////////////////////

LPCSTR __fastcall xmlGetAttr( HXML hXml, int n )
{
	return xi.getAttr( hXml, n );
}

int __fastcall xmlGetAttrCount( HXML hXml )
{
	return xi.getAttrCount( hXml );
}

LPCSTR __fastcall xmlGetAttrName( HXML hXml, int n )
{
	return xi.getAttrName( hXml, n );
}

/////////////////////////////////////////////////////////////////////////////////////////

void __fastcall xmlAddChild( HXML hXml, HXML n, int pos )
{
	xi.addChild2( n, hXml, pos );
}

HXML __fastcall xmlAddChild( HXML hXml, LPCSTR name )
{
	return xi.addChild( hXml, name, NULL );
}

HXML __fastcall xmlAddChild( HXML hXml, LPCSTR name, LPCSTR value )
{
	return xi.addChild( hXml, name, value );
}

HXML __fastcall xmlAddChild( HXML hXml, LPCSTR name, int value )
{
	char buf[40];
	_itoa( value, buf, 10 );
	return xi.addChild( hXml, name, buf );
}

/////////////////////////////////////////////////////////////////////////////////////////

LPCSTR __fastcall xmlGetAttrValue( HXML hXml, LPCSTR key )
{
	return xi.getAttrValue( hXml, key );
}

string __fastcall xmlGetAttrValueString( HXML hXml, LPCSTR key )
{
	LPCSTR pStr = xi.getAttrValue(hXml, key);
	return pStr ? string(pStr) : string();
}

HXML __fastcall xmlGetChild( HXML hXml, int n )
{
	return xi.getChild( hXml, n );
}

HXML __fastcall xmlGetChild( HXML hXml, LPCSTR key )
{
	return xi.getNthChild( hXml, key, 0 );
}

// #if defined( _UNICODE )
// HXML __fastcall xmlGetChild( HXML hXml, LPCSTR key )
// {
// 	HXML result = xi.getNthChild( hXml, key, 0 );
// 	return result;
// }
// 
// HXML __fastcall xmlGetChildByTag( HXML hXml, LPCSTR key, LPCSTR attrName, LPCSTR attrValue )
// {
// 	return xi.getChildByAttrValue( hXml, key, attrName, attrValue );
// }
// #endif

HXML __fastcall xmlGetChildByTag( HXML hXml, LPCSTR key, LPCSTR attrName, LPCSTR attrValue )
{
	HXML result = xi.getChildByAttrValue( hXml, key, attrName, attrValue );
	return result;
}

int __fastcall xmlGetChildCount( HXML hXml )
{
	return xi.getChildCount( hXml );
}

HXML __fastcall xmlGetNthChild( HXML hXml, LPCSTR tag, int nth )
{
	int i, num;

	if ( !hXml || tag == NULL || strlen( tag ) <= 0 || nth < 1 )
		return NULL;

	num = 1;
	for ( i=0; ; i++ ) {
		HXML n = xi.getChild( hXml, i );
		if ( !n )
			break;
		if ( !strcmp( tag, xmlGetName( n ))) {
			if ( num == nth )
				return n;

			num++;
	}	}

	return NULL;
}

LPCSTR __fastcall xmlGetName( HXML xml )
{
	return xi.getName( xml );
}

LPCSTR __fastcall xmlGetText( HXML xml )
{
	return xi.getText( xml );
}

/////////////////////////////////////////////////////////////////////////////////////////

void XPath::ProcessPath(LookupInfo &info, bool bCreate)
{
	if (!info.nodeName) return;

	char *nodeName = (char *)alloca(sizeof(char) * (info.nodeName.length+1));
	lstrcpynA(nodeName, info.nodeName.p, info.nodeName.length+1);

	if (info.attrName && info.attrValue)
	{
		char *attrName = (char *)alloca(sizeof(char) * (info.attrName.length+1));
		lstrcpynA(attrName, info.attrName.p, info.attrName.length+1);
		char *attrValue = (char *)alloca(sizeof(char) * (info.attrValue.length+1));
		lstrcpynA(attrValue, info.attrValue.p, info.attrValue.length+1);
		HXML hXml = xmlGetChildByTag(m_hXml, nodeName, attrName, attrValue);

		m_hXml = (hXml || !bCreate) ? hXml : (m_hXml << XCHILD(nodeName) << XATTR(attrName, attrValue));
	} else
	if (info.nodeIndex)
	{
		int idx = atoi(info.nodeIndex.p);
		HXML hXml = strcmp(nodeName, "*") ? xmlGetNthChild(m_hXml, nodeName, idx) : xmlGetChild(m_hXml, idx-1);
		m_hXml = (hXml || !bCreate) ? hXml : (m_hXml << XCHILD(nodeName));
		// no support for such creation mode
	} else
	{
		HXML hXml = xmlGetChild(m_hXml, nodeName);
		m_hXml = (hXml || !bCreate) ? hXml : (m_hXml << XCHILD(nodeName));
	}

	info.Reset();
}

XPath::PathType XPath::LookupImpl(bool bCreate)
{
	LookupState state = S_START;
	LookupInfo info = {0};

	for (LPCSTR p = m_szPath; state < S_FINAL; ++p)
	{
		switch (state)
		{
		case S_START:
		{
			ProcessPath(info, bCreate);
			if (!m_hXml)
			{
				state = S_FINAL_ERROR;
				break;
			}

			switch (*p)
			{
			case 0:
				state = S_FINAL_ERROR;
				break;
			case '@':
				info.attrName.Begin(p+1);
				state = S_ATTR_STEP;
				break;
			case '/':
				break;
			default:
				info.nodeName.Begin(p);
				state = S_NODE_NAME;
				break;
			};
			break;
		}
		case S_ATTR_STEP:
		{
			switch (*p)
			{
			case 0:
				info.attrName.End(p);
				state = S_FINAL_ATTR;
				break;
			default:
				break;
			};
			break;
		}
		case S_NODE_NAME:
		{
			switch (*p)
			{
			case 0:
				info.nodeName.End(p);
				state = S_FINAL_NODESET;
				break;
			case '[':
				info.nodeName.End(p);
				state = S_NODE_OPENBRACKET;
				break;
			case '/':
				info.nodeName.End(p);
				state = S_START;
				break;
			default:
				break;
			};
			break;
		}
		case S_NODE_OPENBRACKET:
		{
			switch (*p)
			{
			case 0:
				state = S_FINAL_ERROR;
				break;
			case '@':
				info.attrName.Begin(p+1);
				state = S_NODE_ATTRNAME;
				break;
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				info.nodeIndex.Begin(p);
				state = S_NODE_INDEX;
				break;
			default:
				state = S_FINAL_ERROR;
				break;
			};
			break;
		}
		case S_NODE_INDEX:
		{
			switch (*p)
			{
			case 0:
				state = S_FINAL_ERROR;
				break;
			case ']':
				info.nodeIndex.End(p);
				state = S_NODE_CLOSEBRACKET;
				break;
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				break;
			default:
				state = S_FINAL_ERROR;
				break;
			};
			break;
		}
		case S_NODE_ATTRNAME:
		{
			switch (*p)
			{
			case 0:
				state = S_FINAL_ERROR;
				break;
			case '=':
				info.attrName.End(p);
				state = S_NODE_ATTREQUALS;
				break;
			default:
				break;
			};
			break;
		}
		case S_NODE_ATTREQUALS:
		{
			switch (*p)
			{
			case 0:
				state = S_FINAL_ERROR;
				break;
			case '\'':
				info.attrValue.Begin(p+1);
				state = S_NODE_ATTRVALUE;
				break;
			default:
				state = S_FINAL_ERROR;
				break;
			};
			break;
		}
		case S_NODE_ATTRVALUE:
		{
			switch (*p)
			{
			case 0:
				state = S_FINAL_ERROR;
				break;
			case '\'':
				info.attrValue.End(p);
				state = S_NODE_ATTRCLOSEVALUE;
				break;
			default:
				break;
			};
			break;
		}
		case S_NODE_ATTRCLOSEVALUE:
		{
			switch (*p)
			{
			case 0:
				state = S_FINAL_ERROR;
				break;
			case ']':
				state = S_NODE_CLOSEBRACKET;
				break;
			default:
				state = S_FINAL_ERROR;
				break;
			};
			break;
		}
		case S_NODE_CLOSEBRACKET:
		{
			switch (*p)
			{
			case 0:
				state = S_FINAL_NODE;
				break;
			case '/':
				state = S_START;
				break;
			default:
				state = S_FINAL_ERROR;
				break;
			};
			break;
		}
		}

		if (!*p && (state < S_FINAL))
		{
			state = S_FINAL_ERROR;
		}
	}

	switch (state)
	{
	case S_FINAL_ATTR:
		m_szParam = info.attrName.p;
		return T_ATTRIBUTE;
	case S_FINAL_NODE:
		ProcessPath(info, bCreate);
		return T_NODE;
	case S_FINAL_NODESET:
		m_szParam = info.nodeName.p;
		return T_NODESET;
	}

	return T_ERROR;
}
