/*

Jabber Protocol Plugin for Miranda IM
Copyright ( C ) 2002-04  Santithorn Bunchua
Copyright ( C ) 2005-09  George Hazan
Copyright ( C ) 2007     Maxim Mluhov

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or ( at your option ) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

File name      : $URL: https://miranda.svn.sourceforge.net/svnroot/miranda/trunk/miranda/protocols/JabberG/jabber_xml.h $
Revision       : $Revision: 8779 $
Last change on : $Date: 2009-01-07 04:16:38 +0600 (Ср, 07 янв 2009) $
Last change by : $Author: ghazan $

*/

#ifndef _JABBER_XML_H_
#define _JABBER_XML_H_

#include "xml.h"

void    __fastcall xmlAddChild( HXML, HXML, int );
HXML    __fastcall xmlAddChild( HXML, LPCSTR pszName );
HXML    __fastcall xmlAddChild( HXML, LPCSTR pszName, LPCSTR ptszValue );
HXML    __fastcall xmlAddChild( HXML, LPCSTR pszName, int iValue );

//LPCSTR __fastcall xmlGetAttrValue( HXML, LPCSTR key );
HXML    __fastcall xmlGetChild( HXML, int n = 0 );
HXML    __fastcall xmlGetChild( HXML, LPCSTR key );
HXML    __fastcall xmlGetChild( HXML, LPCSTR key );
int     __fastcall xmlGetChildCount( HXML );
HXML    __fastcall xmlGetChildByTag( HXML, LPCSTR key, LPCSTR attrName, LPCSTR attrValue );
HXML    __fastcall xmlGetChildByTag( HXML, LPCSTR key, LPCSTR attrName, LPCSTR attrValue );
HXML    __fastcall xmlGetNthChild( HXML, LPCSTR key, int n = 0 );

LPCSTR __fastcall xmlGetName( HXML );
LPCSTR __fastcall xmlGetText( HXML );

void    __fastcall xmlAddAttr( HXML, LPCSTR pszName, LPCSTR ptszValue );
void    __fastcall xmlAddAttr( HXML, LPCSTR pszName, int value );
void    __fastcall xmlAddAttrID( HXML, int id );

int     __fastcall xmlGetAttrCount( HXML );
LPCSTR __fastcall xmlGetAttr( HXML, int n );
LPCSTR __fastcall xmlGetAttrName( HXML, int n );
LPCSTR __fastcall xmlGetAttrValue( HXML, LPCSTR key );
string __fastcall xmlGetAttrValueString( HXML, LPCSTR key );

struct XmlNode
{
	__forceinline XmlNode() { m_hXml = NULL; }

	__forceinline XmlNode( LPCSTR pszString, int* numBytes, LPCSTR ptszTag )
	{
		m_hXml = xi.parseString( pszString, numBytes, ptszTag );
	}

	XmlNode( const XmlNode& n );
	XmlNode( LPCSTR name );
	XmlNode( LPCSTR pszName, LPCSTR ptszText );
	~XmlNode();

	XmlNode& operator =( const XmlNode& n );

	__forceinline operator HXML() const
	{	return m_hXml;
	}

private:
	HXML m_hXml;
};

/////////////////////////////////////////////////////////////////////////////////////////

struct XATTR
{
	LPCSTR name, value;

	__forceinline XATTR( LPCSTR _name, LPCSTR _value ) :
		name( _name ),
		value( _value )
		{}
};

HXML __forceinline operator<<( HXML node, const XATTR& attr )
{	xmlAddAttr( node, attr.name, attr.value );
	return node;
}

/////////////////////////////////////////////////////////////////////////////////////////

struct XATTRI
{
	LPCSTR name;
	int value;

	__forceinline XATTRI( LPCSTR _name, int _value ) :
		name( _name ),
		value( _value )
		{}
};

HXML __forceinline operator<<( HXML node, const XATTRI& attr )
{	xmlAddAttr( node, attr.name, attr.value );
	return node;
}

/////////////////////////////////////////////////////////////////////////////////////////

struct XATTRID
{
	int id;

	__forceinline XATTRID( int _value ) :
		id( _value )
		{}
};

HXML __forceinline operator<<( HXML node, const XATTRID& attr )
{	xmlAddAttrID( node, attr.id );
	return node;
}

/////////////////////////////////////////////////////////////////////////////////////////

struct XCHILD
{
	LPCSTR name, value;

	__forceinline XCHILD( LPCSTR _name, LPCSTR _value = NULL ) :
		name( _name ),
		value( _value )
		{}
};

HXML __forceinline operator<<( HXML node, const XCHILD& child )
{	return xmlAddChild( node, child.name, child.value );
}

/////////////////////////////////////////////////////////////////////////////////////////

struct XCHILDNS
{
	LPCSTR name, ns;

	__forceinline XCHILDNS( LPCSTR _name, LPCSTR _ns = NULL ) :
		name( _name ),
		ns( _ns )
		{}
};

HXML __fastcall operator<<( HXML node, const XCHILDNS& child );

/////////////////////////////////////////////////////////////////////////////////////////

struct XQUERY
{
	LPCSTR ns;

	__forceinline XQUERY( LPCSTR _ns ) :
		ns( _ns )
		{}
};

HXML __fastcall operator<<( HXML node, const XQUERY& child );

/////////////////////////////////////////////////////////////////////////////////////////
// Limited XPath support
//     path should look like: "node-spec/node-spec/.../result-spec"
//     where "node-spec" can be "node-name", "node-name[@attr-name='attr-value']" or "node-name[node-index]"
//     result may be either "node-spec", or "@attr-name"
//
// Samples:
//    LPCSTR s = XPathT(node, "child/subchild[@attr='value']");          // get node text
//    LPCSTR s = XPathT(node, "child/subchild[2]/@attr");                // get attribute value
//    XPathT(node, "child/subchild[@name='test']/@attr") = _T("Hello");   // create path if needed and set attribute value
//
//    XPathT(node, "child/subchild[@name='test']") = _T("Hello");         // TODO create path if needed and set node text

#define XPathT(a,b) XPath(a, _T(b))

class XPath
{
public:
	__forceinline XPath(HXML hXml, const char *path):
		m_type(T_UNKNOWN),
		m_hXml(hXml),
		m_szPath(strdup(path)),
		m_szParam(NULL)
		{}
	~XPath()
	{
		if (m_szPath)
			free((void*)m_szPath);
	}

	// Read data
	operator HXML()
	{
		switch (Lookup())
		{
			case T_ATTRIBUTE:
			case T_NODE: return m_hXml;
			case T_NODESET: return xmlGetNthChild(m_hXml, m_szParam, 1);
		}
		return NULL;
	}
	operator LPSTR()
	{
		switch (Lookup())
		{
			case T_ATTRIBUTE: return (char *)xmlGetAttrValue(m_hXml, m_szParam);
			case T_NODE: return (char *)xmlGetText(m_hXml);
			case T_NODESET: return (char *)xmlGetText(xmlGetNthChild(m_hXml, m_szParam, 1));
		}
		return NULL;
	}
	operator int()
	{
		if (char *s = *this) return atoi(s);
		return 0;
	}
	__forceinline bool operator== (char *str)
	{
		return !strcmp((LPCSTR)*this, str);
	}
	__forceinline bool operator!= (char *str)
	{
		return strcmp((LPCSTR)*this, str) ? true : false;
	}
	HXML operator[] (int idx)
	{
		return (Lookup() == T_NODESET) ? xmlGetNthChild(m_hXml, m_szParam, idx) : NULL;
	}

	// Write data
	void operator= (LPCSTR value)
	{
		switch (Lookup(true))
		{
			case T_ATTRIBUTE: 
				if (!xi.setAttrValue(m_hXml, m_szParam, value)) {
					xmlAddAttr(m_hXml, m_szParam, value); 
				}
				break;
			case T_NODE: break; // TODO set node text
		}
	}
	void operator= (int value)
	{
		char buf[16];
		itoa(value, buf, 10);
		*this = buf;
	}

//private:
	enum PathType
	{
		T_UNKNOWN,
		T_ERROR,
		T_NODE,
		T_ATTRIBUTE,
		T_NODESET
	};

	__forceinline PathType Lookup(bool bCreate=false)
	{
		return (m_type == T_UNKNOWN) ? m_type = LookupImpl(bCreate) : m_type;
	}

	enum LookupState
	{
		S_START,
		S_ATTR_STEP,
		S_NODE_NAME,
		S_NODE_OPENBRACKET,
		S_NODE_INDEX,
		S_NODE_ATTRNAME,
		S_NODE_ATTREQUALS,
		S_NODE_ATTRVALUE,
		S_NODE_ATTRCLOSEVALUE,
		S_NODE_CLOSEBRACKET,

		S_FINAL,
		S_FINAL_ERROR,
		S_FINAL_ATTR,
		S_FINAL_NODESET,
		S_FINAL_NODE
	};

	struct LookupString
	{
		void Begin(const char *p_) { p = p_; }
		void End(const char *p_) { length = p_ - p; }
		operator bool() { return p ? true : false; }

		const char *p;
		int length;

	};

	struct LookupInfo
	{
		void Reset() { memset(this, 0, sizeof(*this)); }
		LookupString nodeName;
		LookupString attrName;
		LookupString attrValue;
		LookupString nodeIndex;
	};

	void ProcessPath(LookupInfo &info, bool bCreate);
	PathType LookupImpl(bool bCreate);

	PathType m_type;
	HXML m_hXml;
	LPCSTR m_szPath;
	LPCSTR m_szParam;
};

class XPathFmt: public XPath
{
public:
	enum { BUFSIZE = 512 };
	XPathFmt(HXML hXml, char *path, ...): XPath(hXml, m_buf)
	{
		*m_buf = 0;

		va_list args;
		va_start(args, path);
		vsnprintf(m_buf, BUFSIZE, path, args);
		m_buf[BUFSIZE-1] = 0;
		va_end(args);
	}

// #ifdef _UNICODE
// 	XPathFmt(HXML hXml, char *path, ...): XPath(hXml, m_buf)
// 	{
// 		*m_buf = 0;
// 		char buf[BUFSIZE];
// 
// 		va_list args;
// 		va_start(args, path);
// 		_vsnprintf(buf, BUFSIZE, path, args);
// 		buf[BUFSIZE-1] = 0;
// 		MultiByteToWideChar(CP_ACP, 0, buf, -1, m_buf, BUFSIZE);
// 		va_end(args);
// 	}
// #endif

private:
	char m_buf[BUFSIZE];
};

#endif
