/*

Miranda IM: the free IM client for Microsoft* Windows*

Copyright 2000-2009 Miranda ICQ/IM project,
all portions of this codebase are copyrighted to the people
listed in contributors.txt.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "stdafx.h"
#include "xmlParser.h"

static HXML xmlapiCreateNode( LPCSTR name, LPCSTR text, char isDeclaration )
{
	XMLNode result = XMLNode::createXMLTopNode( name, isDeclaration );
	if ( text )
		result.updateText( text );
	return result.detach();
}

static void xmlapiDestroyNode( HXML n )
{
	XMLNode tmp; tmp.attach(n);
}

static HXML xmlapiParseString( LPCSTR str, int* datalen, LPCSTR tag )
{
	XMLResults res;
	XMLNode result = XMLNode::parseString( str, tag, &res );

	if ( datalen != NULL )
		datalen[0] += res.nChars;

	return (tag != NULL || res.error == eXMLErrorNone) ? result.detach() : NULL;
}

static HXML xmlapiAddChild( HXML _n, LPCSTR name, LPCSTR text )
{
	XMLNode result = XMLNode(_n).addChild( name );
	if ( text != NULL )
		result.updateText( text );
	return result;
}

static void xmlapiAddChild2( HXML _child, HXML _parent, int pos)
{
	XMLNode child(_child), parent(_parent);
	parent.addChild( child, pos );
}

static HXML xmlapiCopyNode( HXML _n )
{
	XMLNode result = XMLNode(_n);
	return result.detach();
}

static LPCSTR xmlapiGetAttr( HXML _n, int i )
{
	return XMLNode(_n).getAttributeValue( i );
}

static int xmlapiGetAttrCount( HXML _n )
{
	return XMLNode(_n).nAttribute();
}

static LPCSTR xmlapiGetAttrName( HXML _n, int i )
{
	return XMLNode(_n).getAttributeName( i );
}

static HXML xmlapiGetChild( HXML _n, int i )
{
	return XMLNode(_n).getChildNode( i );
}

static HXML xmlapiGetChildByAttrValue( HXML _n, LPCSTR name, LPCSTR attrName, LPCSTR attrValue )
{
	return XMLNode(_n).getChildNodeWithAttribute( name, attrName, attrValue );
}

static int xmlapiGetChildCount( HXML _n )
{
	return XMLNode(_n).nChildNode();
}

static HXML xmlapiGetFirstChild( HXML _n )
{
	return XMLNode(_n).getChildNode( 0 );
}

static HXML xmlapiGetNthChild( HXML _n, LPCSTR name, int i )
{
	return XMLNode(_n).getChildNode( name, i );
}

static HXML xmlapiGetNextChild( HXML _n, LPCSTR name, int* i )
{
	return XMLNode(_n).getChildNode( name, i );
}

static HXML xmlapiGetNextNode( HXML _n )
{
	return XMLNode(_n).getNextNode( );
}

static HXML xmlapiGetChildByPath( HXML _n, LPCSTR path, char createNodeIfMissing )
{
	return XMLNode(_n).getChildNodeByPath( path, createNodeIfMissing );
}

static LPCSTR xmlapiGetName( HXML _n )
{
	return XMLNode(_n).getName();
}

static HXML xmlapiGetParent( HXML _n )
{
	return XMLNode(_n).getParentNode();
}

static LPCSTR xmlapiGetText( HXML _n )
{
	return XMLNode(_n).getText();
}

static LPCSTR xmlapiGetAttrValue( HXML _n, LPCSTR attrName )
{
	return XMLNode(_n).getAttribute( attrName );
}
static int xmlapiSetAttribute( HXML _n, LPCSTR attrName,  XMLCSTR val)
{
	return XMLNode(_n).setAttribute( attrName, val );
}
static void xmlapiSetText( HXML _n, LPCSTR _text )
{
	XMLNode(_n).updateText( _text );
}

static LPSTR xmlapiToString( HXML _n, int* datalen )
{
	return XMLNode(_n).createXMLString( 0, datalen );
}

static LPSTR xmlapiToStringF( HXML _n, int* datalen )
{
	return XMLNode(_n).createXMLStringF( datalen );
}

static void xmlapiAddAttr( HXML _n, LPCSTR attrName, LPCSTR attrValue )
{
	if ( attrName != NULL && attrValue != NULL )
		XMLNode(_n).addAttribute( attrName, attrValue );
}

static void xmlapiAddAttrInt( HXML _n, LPCSTR attrName, int attrValue )
{
	char buf[40];
	_itoa( attrValue, buf, 10 );
	XMLNode(_n).addAttribute( attrName, buf );
}

static void xmlapiFree( void* p )
{
	free( p );
}

/////////////////////////////////////////////////////////////////////////////////////////

bool GetXmlApi(XML_API* xi)
{
	if ( xi == NULL )
		return false;

	if ( xi->cbSize != sizeof(XML_API))
		return false;

	xi->createNode          = xmlapiCreateNode;
	xi->destroyNode         = xmlapiDestroyNode;

	xi->parseString         = xmlapiParseString;
	xi->toString            = xmlapiToString;
	xi->toStringF           = xmlapiToStringF;
	xi->freeMem             = xmlapiFree;

	xi->addChild            = xmlapiAddChild;
	xi->addChild2           = xmlapiAddChild2;
	xi->copyNode            = xmlapiCopyNode;
	xi->getChild            = xmlapiGetChild;
	xi->getChildByAttrValue = xmlapiGetChildByAttrValue;
	xi->getChildCount       = xmlapiGetChildCount;
	xi->getFirstChild       = xmlapiGetFirstChild;
	xi->getNthChild         = xmlapiGetNthChild;
	xi->getNextChild        = xmlapiGetNextChild;
	xi->getNextNode         = xmlapiGetNextNode;
	xi->getChildByPath      = xmlapiGetChildByPath;
	xi->getName             = xmlapiGetName;
	xi->getParent           = xmlapiGetParent;
	xi->getText             = xmlapiGetText;
	xi->setText             = xmlapiSetText;

	xi->getAttr             = xmlapiGetAttr;
	xi->getAttrCount        = xmlapiGetAttrCount;
	xi->getAttrName         = xmlapiGetAttrName;
	xi->getAttrValue        = xmlapiGetAttrValue;
	xi->setAttrValue        = xmlapiSetAttribute;
	xi->addAttr             = xmlapiAddAttr;
	xi->addAttrInt          = xmlapiAddAttrInt;
	return false;
}

// void InitXmlApi( void )
// {
// 	CreateServiceFunction( MS_SYSTEM_GET_XI, GetXmlApi );
// }
